'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _util = require('util');

var _util2 = _interopRequireDefault(_util);

var _v = require('uuid/v4');

var _v2 = _interopRequireDefault(_v);

var _redisApi = require('./redisApi');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

var Z_VALUE_SEPARATOR = '' + String.fromCharCode(0x0) + String.fromCharCode(0x0);
var Z_LEX_MAX_VALUE = String.fromCharCode(0xff);
var TTL_FOR_TEMP_COLLECTION = 60;

var getDocument = function () {
    var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(_ref2, collectionName, id) {
        var client = _ref2.client;
        return regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
                switch (_context.prev = _context.next) {
                    case 0:
                        _context.next = 2;
                        return (0, _redisApi.hget)(client, collectionName, id);

                    case 2:
                        return _context.abrupt('return', _context.sent);

                    case 3:
                    case 'end':
                        return _context.stop();
                }
            }
        }, _callee, undefined);
    }));

    return function getDocument(_x, _x2, _x3) {
        return _ref.apply(this, arguments);
    };
}();

var validateIndexes = function validateIndexes(indexes, fieldNames, errorMessage) {
    fieldNames.forEach(function (field) {
        if (!indexes[field]) {
            throw new Error(_util2.default.format(errorMessage, field));
        }
    });
};

var validateCriteriaFields = function validateCriteriaFields(indexes, criteria) {
    if (criteria) {
        validateIndexes(indexes, Object.keys(criteria), 'Can\'t find index for \'%s\': you can find by only indexed field');
    }
};

var saveIdsToCollection = function () {
    var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(_ref4, collectionName, ids) {
        var client = _ref4.client;
        var args;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
            while (1) {
                switch (_context2.prev = _context2.next) {
                    case 0:
                        args = ids.reduce(function (acc, id) {
                            // score
                            acc.push(id);
                            // value
                            acc.push(id);
                            return acc;
                        }, [collectionName]);
                        _context2.next = 3;
                        return (0, _redisApi.zaddMulti)(client, args);

                    case 3:
                        _context2.next = 5;
                        return (0, _redisApi.expire)(client, collectionName, TTL_FOR_TEMP_COLLECTION);

                    case 5:
                    case 'end':
                        return _context2.stop();
                }
            }
        }, _callee2, undefined);
    }));

    return function saveIdsToCollection(_x4, _x5, _x6) {
        return _ref3.apply(this, arguments);
    };
}();

var getIds = function () {
    var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(repository, collectionName, indexes, criteria) {
        var client, metaCollection, tempPrefix, tempCollectionNames, promises, keys, result;
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
            while (1) {
                switch (_context4.prev = _context4.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;


                        validateCriteriaFields(indexes, criteria);
                        tempPrefix = (0, _v2.default)();
                        tempCollectionNames = {};

                        // todo extract function

                        promises = Object.keys(criteria).map(function () {
                            var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3(fieldName) {
                                var fieldValue, indexName, fieldType, ids, lexIds;
                                return regeneratorRuntime.wrap(function _callee3$(_context3) {
                                    while (1) {
                                        switch (_context3.prev = _context3.next) {
                                            case 0:
                                                tempCollectionNames[fieldName] = tempPrefix + '_' + fieldName;
                                                fieldValue = criteria[fieldName];
                                                indexName = metaCollection.getIndexName(collectionName, fieldName);
                                                fieldType = indexes[fieldName].fieldType;
                                                ids = [];

                                                if (!(fieldType === 'number')) {
                                                    _context3.next = 11;
                                                    break;
                                                }

                                                _context3.next = 8;
                                                return (0, _redisApi.zrangebyscore)(client, indexName, fieldValue, fieldValue);

                                            case 8:
                                                ids = _context3.sent;
                                                _context3.next = 15;
                                                break;

                                            case 11:
                                                _context3.next = 13;
                                                return (0, _redisApi.zrangebylex)(client, indexName, '[' + fieldValue + Z_VALUE_SEPARATOR, '(' + fieldValue + Z_VALUE_SEPARATOR + Z_LEX_MAX_VALUE);

                                            case 13:
                                                lexIds = _context3.sent;

                                                ids = lexIds.map(function (value) {
                                                    return value.substr(value.lastIndexOf(Z_VALUE_SEPARATOR) + Z_VALUE_SEPARATOR.length);
                                                });

                                            case 15:
                                                _context3.next = 17;
                                                return saveIdsToCollection(repository, tempCollectionNames[fieldName], ids);

                                            case 17:
                                                return _context3.abrupt('return', _context3.sent);

                                            case 18:
                                            case 'end':
                                                return _context3.stop();
                                        }
                                    }
                                }, _callee3, undefined);
                            }));

                            return function (_x11) {
                                return _ref6.apply(this, arguments);
                            };
                        }());
                        _context4.next = 7;
                        return Promise.all(promises);

                    case 7:
                        keys = Object.values(tempCollectionNames);
                        _context4.next = 10;
                        return (0, _redisApi.zinterstore)(client, tempPrefix, keys.length, keys);

                    case 10:
                        _context4.next = 12;
                        return (0, _redisApi.expire)(client, tempPrefix, TTL_FOR_TEMP_COLLECTION);

                    case 12:
                        _context4.next = 14;
                        return (0, _redisApi.zrange)(client, tempPrefix);

                    case 14:
                        result = _context4.sent;
                        _context4.next = 17;
                        return _redisApi.del.apply(undefined, [client].concat(_toConsumableArray(keys.concat(tempPrefix))));

                    case 17:
                        return _context4.abrupt('return', result);

                    case 18:
                    case 'end':
                        return _context4.stop();
                }
            }
        }, _callee4, undefined);
    }));

    return function getIds(_x7, _x8, _x9, _x10) {
        return _ref5.apply(this, arguments);
    };
}();

var populateNumberIndex = function () {
    var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(client, collectionName, score, id) {
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
            while (1) {
                switch (_context5.prev = _context5.next) {
                    case 0:
                        _context5.next = 2;
                        return (0, _redisApi.zadd)(client, collectionName, score, id);

                    case 2:
                    case 'end':
                        return _context5.stop();
                }
            }
        }, _callee5, undefined);
    }));

    return function populateNumberIndex(_x12, _x13, _x14, _x15) {
        return _ref7.apply(this, arguments);
    };
}();

var populateStringIndex = function () {
    var _ref8 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(client, collectionName, value, id) {
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
            while (1) {
                switch (_context6.prev = _context6.next) {
                    case 0:
                        _context6.next = 2;
                        return (0, _redisApi.zadd)(client, collectionName, 0, '' + value + Z_VALUE_SEPARATOR + id);

                    case 2:
                    case 'end':
                        return _context6.stop();
                }
            }
        }, _callee6, undefined);
    }));

    return function populateStringIndex(_x16, _x17, _x18, _x19) {
        return _ref8.apply(this, arguments);
    };
}();

var updateIndex = function () {
    var _ref9 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(_ref10, collectionName, id, document, _ref11) {
        var client = _ref10.client,
            metaCollection = _ref10.metaCollection;
        var fieldName = _ref11.fieldName,
            fieldType = _ref11.fieldType;
        var indexCollectionName, value;
        return regeneratorRuntime.wrap(function _callee7$(_context7) {
            while (1) {
                switch (_context7.prev = _context7.next) {
                    case 0:
                        indexCollectionName = metaCollection.getIndexName(collectionName, fieldName);
                        value = document[fieldName];

                        if (!(fieldType === 'number')) {
                            _context7.next = 6;
                            break;
                        }

                        if (!(typeof value !== 'number')) {
                            _context7.next = 5;
                            break;
                        }

                        throw new Error('Can\'t update index \'' + fieldName + '\' value: Value \'' + value + '\' is not number.');

                    case 5:
                        return _context7.abrupt('return', populateNumberIndex(client, indexCollectionName, value, id));

                    case 6:
                        return _context7.abrupt('return', populateStringIndex(client, indexCollectionName, value, id));

                    case 7:
                    case 'end':
                        return _context7.stop();
                }
            }
        }, _callee7, undefined);
    }));

    return function updateIndex(_x20, _x21, _x22, _x23, _x24) {
        return _ref9.apply(this, arguments);
    };
}();

var updateIndexes = function () {
    var _ref12 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee9(repository, collectionName, id, document) {
        var metaCollection, indexes, promises;
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
            while (1) {
                switch (_context9.prev = _context9.next) {
                    case 0:
                        metaCollection = repository.metaCollection;
                        _context9.next = 3;
                        return metaCollection.getIndexes(collectionName);

                    case 3:
                        indexes = _context9.sent;
                        promises = Object.keys(indexes).map(function () {
                            var _ref13 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8(fieldName) {
                                var index;
                                return regeneratorRuntime.wrap(function _callee8$(_context8) {
                                    while (1) {
                                        switch (_context8.prev = _context8.next) {
                                            case 0:
                                                index = indexes[fieldName];
                                                _context8.next = 3;
                                                return updateIndex(repository, collectionName, id, document, index);

                                            case 3:
                                            case 'end':
                                                return _context8.stop();
                                        }
                                    }
                                }, _callee8, undefined);
                            }));

                            return function (_x29) {
                                return _ref13.apply(this, arguments);
                            };
                        }());
                        _context9.next = 7;
                        return Promise.all(promises);

                    case 7:
                    case 'end':
                        return _context9.stop();
                }
            }
        }, _callee9, undefined);
    }));

    return function updateIndexes(_x25, _x26, _x27, _x28) {
        return _ref12.apply(this, arguments);
    };
}();

var criteriaIsEmpty = function criteriaIsEmpty(criteria) {
    return !(criteria && Object.keys(criteria).length);
};

var count = function () {
    var _ref14 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee10(_ref15, collectionName, criteria) {
        var client = _ref15.client;
        return regeneratorRuntime.wrap(function _callee10$(_context10) {
            while (1) {
                switch (_context10.prev = _context10.next) {
                    case 0:
                        _context10.next = 2;
                        return (0, _redisApi.hlen)(client, collectionName);

                    case 2:
                        return _context10.abrupt('return', _context10.sent);

                    case 3:
                    case 'end':
                        return _context10.stop();
                }
            }
        }, _callee10, undefined);
    }));

    return function count(_x30, _x31, _x32) {
        return _ref14.apply(this, arguments);
    };
}();

var validateSortFields = function validateSortFields(indexes, sort) {
    if (sort) {
        validateIndexes(indexes, Object.keys(sort), 'Can\'t find index for \'%s\': you can sort by only indexed field');
    }
};

var find = function () {
    var _ref16 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee12(repository, collectionName, criteria, _ref17) {
        var _ref17$skip = _ref17.skip,
            skip = _ref17$skip === undefined ? null : _ref17$skip,
            _ref17$limit = _ref17.limit,
            limit = _ref17$limit === undefined ? null : _ref17$limit,
            _ref17$sort = _ref17.sort,
            sort = _ref17$sort === undefined ? null : _ref17$sort;
        var client, metaCollection, indexes, ids, promises;
        return regeneratorRuntime.wrap(function _callee12$(_context12) {
            while (1) {
                switch (_context12.prev = _context12.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;
                        _context12.next = 3;
                        return metaCollection.getIndexes(collectionName);

                    case 3:
                        indexes = _context12.sent;

                        validateCriteriaFields(indexes, criteria);
                        validateSortFields(indexes, sort);

                        if (!criteriaIsEmpty(criteria)) {
                            _context12.next = 10;
                            break;
                        }

                        _context12.next = 9;
                        return (0, _redisApi.hvals)(client, collectionName);

                    case 9:
                        return _context12.abrupt('return', _context12.sent);

                    case 10:
                        _context12.next = 12;
                        return getIds(repository, collectionName, indexes, criteria);

                    case 12:
                        ids = _context12.sent;

                        // sort
                        // get by ids
                        // promises = ids.map
                        promises = ids.map(function () {
                            var _ref18 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee11(id) {
                                return regeneratorRuntime.wrap(function _callee11$(_context11) {
                                    while (1) {
                                        switch (_context11.prev = _context11.next) {
                                            case 0:
                                                _context11.next = 2;
                                                return getDocument(repository, collectionName, id);

                                            case 2:
                                                return _context11.abrupt('return', _context11.sent);

                                            case 3:
                                            case 'end':
                                                return _context11.stop();
                                        }
                                    }
                                }, _callee11, undefined);
                            }));

                            return function (_x37) {
                                return _ref18.apply(this, arguments);
                            };
                        }());
                        _context12.next = 16;
                        return Promise.all(promises);

                    case 16:
                        return _context12.abrupt('return', _context12.sent);

                    case 17:
                    case 'end':
                        return _context12.stop();
                }
            }
        }, _callee12, undefined);
    }));

    return function find(_x33, _x34, _x35, _x36) {
        return _ref16.apply(this, arguments);
    };
}();

var findOne = function () {
    var _ref19 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee13(repository, collectionName, criteria, _ref20) {
        var _ref20$skip = _ref20.skip,
            skip = _ref20$skip === undefined ? 0 : _ref20$skip,
            _ref20$order = _ref20.order,
            order = _ref20$order === undefined ? 0 : _ref20$order;
        var limit;
        return regeneratorRuntime.wrap(function _callee13$(_context13) {
            while (1) {
                switch (_context13.prev = _context13.next) {
                    case 0:
                        limit = 1;
                        return _context13.abrupt('return', find(repository, collectionName, criteria, { skip: skip, limit: limit, order: order }));

                    case 2:
                    case 'end':
                        return _context13.stop();
                }
            }
        }, _callee13, undefined);
    }));

    return function findOne(_x38, _x39, _x40, _x41) {
        return _ref19.apply(this, arguments);
    };
}();

function wrapFind(repository, initialFind, collectionName, expression) {
    var resultPromise = Promise.resolve();
    var requestChain = {
        skip: null,
        limit: null,
        sort: null
    };
    var sanitizeError = ''; //sanitizeSearchExpression(expression);

    ['skip', 'limit', 'sort'].forEach(function (cmd) {
        resultPromise[cmd] = function (value) {
            requestChain[cmd] = value;
            return resultPromise;
        };
    });

    var originalThen = resultPromise.then.bind(resultPromise);
    var boundExecFind = (!sanitizeError ? initialFind : Promise.reject.bind(Promise, new Error(sanitizeError))).bind(null, repository, collectionName, expression, requestChain);

    resultPromise.then = function () {
        var _originalThen;

        return (_originalThen = originalThen(boundExecFind)).then.apply(_originalThen, arguments);
    };

    resultPromise.catch = function () {
        var _originalThen2;

        return (_originalThen2 = originalThen(boundExecFind)).catch.apply(_originalThen2, arguments);
    };

    return resultPromise;
}

var insert = function () {
    var _ref21 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee14(repository, collectionName, document) {
        var client, metaCollection, _id, member;

        return regeneratorRuntime.wrap(function _callee14$(_context14) {
            while (1) {
                switch (_context14.prev = _context14.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;
                        _context14.next = 3;
                        return metaCollection.getNextId(collectionName);

                    case 3:
                        _id = _context14.sent;
                        member = _extends({ _id: _id }, document);
                        _context14.next = 7;
                        return (0, _redisApi.hset)(client, collectionName, _id, member);

                    case 7:
                        _context14.next = 9;
                        return updateIndexes(repository, collectionName, _id, member);

                    case 9:
                    case 'end':
                        return _context14.stop();
                }
            }
        }, _callee14, undefined);
    }));

    return function insert(_x42, _x43, _x44) {
        return _ref21.apply(this, arguments);
    };
}();

var removeAll = function () {
    var _ref22 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee16(_ref23, collectionName) {
        var client = _ref23.client,
            metaCollection = _ref23.metaCollection;
        var indexes;
        return regeneratorRuntime.wrap(function _callee16$(_context16) {
            while (1) {
                switch (_context16.prev = _context16.next) {
                    case 0:
                        _context16.next = 2;
                        return metaCollection.getIndexes(collectionName);

                    case 2:
                        indexes = _context16.sent;

                        Object.keys(indexes).forEach(function () {
                            var _ref24 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee15(key) {
                                var fieldName, indexCollectionName;
                                return regeneratorRuntime.wrap(function _callee15$(_context15) {
                                    while (1) {
                                        switch (_context15.prev = _context15.next) {
                                            case 0:
                                                fieldName = indexes[key].fieldName;
                                                indexCollectionName = metaCollection.getIndexName(collectionName, fieldName);
                                                _context15.next = 4;
                                                return (0, _redisApi.del)(client, indexCollectionName);

                                            case 4:
                                            case 'end':
                                                return _context15.stop();
                                        }
                                    }
                                }, _callee15, undefined);
                            }));

                            return function (_x47) {
                                return _ref24.apply(this, arguments);
                            };
                        }());
                        _context16.next = 6;
                        return (0, _redisApi.del)(client, collectionName);

                    case 6:
                    case 'end':
                        return _context16.stop();
                }
            }
        }, _callee16, undefined);
    }));

    return function removeAll(_x45, _x46) {
        return _ref22.apply(this, arguments);
    };
}();

var removeIdFromIndex = function () {
    var _ref25 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee17(_ref26, collectionName, id, document, index) {
        var client = _ref26.client,
            metaCollection = _ref26.metaCollection;
        var fieldName, fieldType, indexCollectionName, fieldValue;
        return regeneratorRuntime.wrap(function _callee17$(_context17) {
            while (1) {
                switch (_context17.prev = _context17.next) {
                    case 0:
                        fieldName = index.fieldName, fieldType = index.fieldType;
                        indexCollectionName = metaCollection.getIndexName(collectionName, fieldName);

                        if (!(fieldType === 'number')) {
                            _context17.next = 6;
                            break;
                        }

                        _context17.next = 5;
                        return (0, _redisApi.zrem)(client, indexCollectionName, id);

                    case 5:
                        return _context17.abrupt('return', _context17.sent);

                    case 6:
                        fieldValue = document[fieldName];
                        _context17.next = 9;
                        return (0, _redisApi.zremrangebylex)(client, indexCollectionName, '[' + fieldValue + Z_VALUE_SEPARATOR + id, '(' + fieldValue + Z_VALUE_SEPARATOR + id);

                    case 9:
                    case 'end':
                        return _context17.stop();
                }
            }
        }, _callee17, undefined);
    }));

    return function removeIdFromIndex(_x48, _x49, _x50, _x51, _x52) {
        return _ref25.apply(this, arguments);
    };
}();

var removeIdsFromIndexes = function () {
    var _ref27 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee20(repository, collectionName, indexes, ids) {
        var promises;
        return regeneratorRuntime.wrap(function _callee20$(_context20) {
            while (1) {
                switch (_context20.prev = _context20.next) {
                    case 0:
                        promises = ids.map(function () {
                            var _ref28 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee19(id) {
                                var document, proms;
                                return regeneratorRuntime.wrap(function _callee19$(_context19) {
                                    while (1) {
                                        switch (_context19.prev = _context19.next) {
                                            case 0:
                                                _context19.next = 2;
                                                return getDocument(repository, collectionName, id);

                                            case 2:
                                                document = _context19.sent;
                                                proms = Object.keys(indexes).map(function () {
                                                    var _ref29 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee18(key) {
                                                        return regeneratorRuntime.wrap(function _callee18$(_context18) {
                                                            while (1) {
                                                                switch (_context18.prev = _context18.next) {
                                                                    case 0:
                                                                        _context18.next = 2;
                                                                        return removeIdFromIndex(repository, collectionName, id, document, indexes[key]);

                                                                    case 2:
                                                                        return _context18.abrupt('return', _context18.sent);

                                                                    case 3:
                                                                    case 'end':
                                                                        return _context18.stop();
                                                                }
                                                            }
                                                        }, _callee18, undefined);
                                                    }));

                                                    return function (_x58) {
                                                        return _ref29.apply(this, arguments);
                                                    };
                                                }());
                                                _context19.next = 6;
                                                return Promise.all(proms);

                                            case 6:
                                            case 'end':
                                                return _context19.stop();
                                        }
                                    }
                                }, _callee19, undefined);
                            }));

                            return function (_x57) {
                                return _ref28.apply(this, arguments);
                            };
                        }());
                        _context20.next = 3;
                        return Promise.all(promises);

                    case 3:
                    case 'end':
                        return _context20.stop();
                }
            }
        }, _callee20, undefined);
    }));

    return function removeIdsFromIndexes(_x53, _x54, _x55, _x56) {
        return _ref27.apply(this, arguments);
    };
}();

var removeIds = function () {
    var _ref30 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee21(repository, collectionName, indexes, ids) {
        return regeneratorRuntime.wrap(function _callee21$(_context21) {
            while (1) {
                switch (_context21.prev = _context21.next) {
                    case 0:
                        _context21.next = 2;
                        return removeIdsFromIndexes(repository, collectionName, indexes, ids);

                    case 2:
                        _context21.next = 4;
                        return _redisApi.hdel.apply(undefined, [repository.client, collectionName].concat(_toConsumableArray(ids)));

                    case 4:
                    case 'end':
                        return _context21.stop();
                }
            }
        }, _callee21, undefined);
    }));

    return function removeIds(_x59, _x60, _x61, _x62) {
        return _ref30.apply(this, arguments);
    };
}();

var remove = function () {
    var _ref31 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee22(repository, collectionName, criteria) {
        var metaCollection, indexes, ids;
        return regeneratorRuntime.wrap(function _callee22$(_context22) {
            while (1) {
                switch (_context22.prev = _context22.next) {
                    case 0:
                        metaCollection = repository.metaCollection;

                        if (!(!criteria || !Object.keys(criteria).length)) {
                            _context22.next = 5;
                            break;
                        }

                        _context22.next = 4;
                        return removeAll(repository, collectionName);

                    case 4:
                        return _context22.abrupt('return', _context22.sent);

                    case 5:
                        _context22.next = 7;
                        return metaCollection.getIndexes(collectionName);

                    case 7:
                        indexes = _context22.sent;
                        _context22.next = 10;
                        return getIds(repository, collectionName, indexes, criteria);

                    case 10:
                        ids = _context22.sent;
                        _context22.next = 13;
                        return removeIds(repository, collectionName, indexes, ids);

                    case 13:
                    case 'end':
                        return _context22.stop();
                }
            }
        }, _callee22, undefined);
    }));

    return function remove(_x63, _x64, _x65) {
        return _ref31.apply(this, arguments);
    };
}();

var applyDotNotation = function () {
    var _ref32 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee23(document, fieldName, fieldOptions, cb) {
        var fields;
        return regeneratorRuntime.wrap(function _callee23$(_context23) {
            while (1) {
                switch (_context23.prev = _context23.next) {
                    case 0:
                        fields = fieldName.split('.');

                        if (fields.length > 1) {
                            fields.reduce(function (acc, field, currentIndex) {
                                if (fields.length === currentIndex - 1) {
                                    cb(acc, fieldName, fieldOptions, false);
                                }
                                return acc[fieldName];
                            }, document);
                        } else {
                            cb(document, fieldName, fieldOptions, true);
                        }

                    case 2:
                    case 'end':
                        return _context23.stop();
                }
            }
        }, _callee23, undefined);
    }));

    return function applyDotNotation(_x66, _x67, _x68, _x69) {
        return _ref32.apply(this, arguments);
    };
}();

var applyFields = function () {
    var _ref33 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee25(document, options, cb) {
        var promises;
        return regeneratorRuntime.wrap(function _callee25$(_context25) {
            while (1) {
                switch (_context25.prev = _context25.next) {
                    case 0:
                        promises = Object.keys(options).map(function () {
                            var _ref34 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee24(fieldName) {
                                return regeneratorRuntime.wrap(function _callee24$(_context24) {
                                    while (1) {
                                        switch (_context24.prev = _context24.next) {
                                            case 0:
                                                return _context24.abrupt('return', applyDotNotation(document, fieldName, options[fieldName], cb));

                                            case 1:
                                            case 'end':
                                                return _context24.stop();
                                        }
                                    }
                                }, _callee24, undefined);
                            }));

                            return function (_x73) {
                                return _ref34.apply(this, arguments);
                            };
                        }());
                        _context25.next = 3;
                        return Promise.all(promises);

                    case 3:
                    case 'end':
                        return _context25.stop();
                }
            }
        }, _callee25, undefined);
    }));

    return function applyFields(_x70, _x71, _x72) {
        return _ref33.apply(this, arguments);
    };
}();

var updateHandlers = {
    $unset: function () {
        var _ref35 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee26(repository, collectionName, indexes, id, partDoc, fieldName, isRootLevel) {
            var index;
            return regeneratorRuntime.wrap(function _callee26$(_context26) {
                while (1) {
                    switch (_context26.prev = _context26.next) {
                        case 0:
                            if (!isRootLevel) {
                                _context26.next = 5;
                                break;
                            }

                            index = indexes[fieldName];

                            if (!index) {
                                _context26.next = 5;
                                break;
                            }

                            _context26.next = 5;
                            return removeIdFromIndex(repository, collectionName, id, partDoc, index);

                        case 5:

                            if (Array.isArray(partDoc[fieldName])) {
                                partDoc[fieldName] = null;
                            } else {
                                delete partDoc[fieldName];
                            }

                        case 6:
                        case 'end':
                            return _context26.stop();
                    }
                }
            }, _callee26, undefined);
        }));

        return function $unset(_x74, _x75, _x76, _x77, _x78, _x79, _x80) {
            return _ref35.apply(this, arguments);
        };
    }(),
    $inc: function () {
        var _ref36 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee27(repository, collectionName, indexes, id, partDoc, fieldName, isRootLevel, fieldOptions) {
            var index, _index;

            return regeneratorRuntime.wrap(function _callee27$(_context27) {
                while (1) {
                    switch (_context27.prev = _context27.next) {
                        case 0:
                            if (!isRootLevel) {
                                _context27.next = 5;
                                break;
                            }

                            index = indexes[fieldName];

                            if (!index) {
                                _context27.next = 5;
                                break;
                            }

                            _context27.next = 5;
                            return removeIdFromIndex(repository, collectionName, id, partDoc, index);

                        case 5:

                            partDoc[fieldName] += fieldOptions;

                            if (!isRootLevel) {
                                _context27.next = 11;
                                break;
                            }

                            _index = indexes[fieldName];

                            if (!_index) {
                                _context27.next = 11;
                                break;
                            }

                            _context27.next = 11;
                            return updateIndex(repository, collectionName, id, document, _index);

                        case 11:
                        case 'end':
                            return _context27.stop();
                    }
                }
            }, _callee27, undefined);
        }));

        return function $inc(_x81, _x82, _x83, _x84, _x85, _x86, _x87, _x88) {
            return _ref36.apply(this, arguments);
        };
    }(),
    $push: function () {
        var _ref37 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee28(repository, collectionName, indexes, id, partDoc, fieldName, isRootLevel, fieldOptions) {
            var index, _index2;

            return regeneratorRuntime.wrap(function _callee28$(_context28) {
                while (1) {
                    switch (_context28.prev = _context28.next) {
                        case 0:
                            if (Array.isArray(partDoc[fieldName])) {
                                _context28.next = 2;
                                break;
                            }

                            throw new Error('The field \'' + fieldName + '\' is not array in document { _id: ' + id + ' }');

                        case 2:
                            if (!isRootLevel) {
                                _context28.next = 7;
                                break;
                            }

                            index = indexes[fieldName];

                            if (!index) {
                                _context28.next = 7;
                                break;
                            }

                            _context28.next = 7;
                            return removeIdFromIndex(repository, collectionName, id, partDoc, index);

                        case 7:
                            document[fieldName] = fieldOptions;

                            if (!isRootLevel) {
                                _context28.next = 13;
                                break;
                            }

                            _index2 = indexes[fieldName];

                            if (!_index2) {
                                _context28.next = 13;
                                break;
                            }

                            _context28.next = 13;
                            return updateIndex(repository, collectionName, id, document, _index2);

                        case 13:
                        case 'end':
                            return _context28.stop();
                    }
                }
            }, _callee28, undefined);
        }));

        return function $push(_x89, _x90, _x91, _x92, _x93, _x94, _x95, _x96) {
            return _ref37.apply(this, arguments);
        };
    }(),
    $pull: function () {
        var _ref38 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee29(repository, collectionName, indexes, id, document, options) {
            return regeneratorRuntime.wrap(function _callee29$(_context29) {
                while (1) {
                    switch (_context29.prev = _context29.next) {
                        case 0:
                            throw new Error('Not implemented!!!');

                        case 1:
                        case 'end':
                            return _context29.stop();
                    }
                }
            }, _callee29, undefined);
        }));

        return function $pull(_x97, _x98, _x99, _x100, _x101, _x102) {
            return _ref38.apply(this, arguments);
        };
    }()
};

var update = function () {
    var _ref39 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee32(repository, collectionName, criteria, operators) {
        var client, metaCollection, indexes, ids;
        return regeneratorRuntime.wrap(function _callee32$(_context32) {
            while (1) {
                switch (_context32.prev = _context32.next) {
                    case 0:
                        client = repository.client, metaCollection = repository.metaCollection;
                        _context32.next = 3;
                        return metaCollection.getIndexes(collectionName);

                    case 3:
                        indexes = _context32.sent;
                        _context32.next = 6;
                        return getIds(repository, collectionName, indexes, criteria);

                    case 6:
                        ids = _context32.sent;

                        ids.forEach(function () {
                            var _ref40 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee31(id) {
                                var document, promises;
                                return regeneratorRuntime.wrap(function _callee31$(_context31) {
                                    while (1) {
                                        switch (_context31.prev = _context31.next) {
                                            case 0:
                                                _context31.next = 2;
                                                return getDocument(repository, collectionName, id);

                                            case 2:
                                                document = _context31.sent;
                                                promises = Object.keys(operators).map(function () {
                                                    var _ref41 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee30(operatorName) {
                                                        var handler, operatorOptions;
                                                        return regeneratorRuntime.wrap(function _callee30$(_context30) {
                                                            while (1) {
                                                                switch (_context30.prev = _context30.next) {
                                                                    case 0:
                                                                        handler = updateHandlers[operatorName];

                                                                        if (handler) {
                                                                            _context30.next = 3;
                                                                            break;
                                                                        }

                                                                        throw new Error('Invalid update operator');

                                                                    case 3:
                                                                        operatorOptions = operators[operatorName];
                                                                        _context30.next = 6;
                                                                        return applyFields(document, operatorOptions, function (doc, fieldName, fieldOptions, isRootLevel) {
                                                                            handler(repository, collectionName, indexes, id, doc, fieldName, isRootLevel, fieldOptions);
                                                                        });

                                                                    case 6:
                                                                    case 'end':
                                                                        return _context30.stop();
                                                                }
                                                            }
                                                        }, _callee30, undefined);
                                                    }));

                                                    return function (_x108) {
                                                        return _ref41.apply(this, arguments);
                                                    };
                                                }());
                                                _context31.next = 6;
                                                return Promise.all(promises);

                                            case 6:
                                                _context31.next = 8;
                                                return (0, _redisApi.hset)(client, collectionName, id, document);

                                            case 8:
                                            case 'end':
                                                return _context31.stop();
                                        }
                                    }
                                }, _callee31, undefined);
                            }));

                            return function (_x107) {
                                return _ref40.apply(this, arguments);
                            };
                        }());

                    case 8:
                    case 'end':
                        return _context32.stop();
                }
            }
        }, _callee32, undefined);
    }));

    return function update(_x103, _x104, _x105, _x106) {
        return _ref39.apply(this, arguments);
    };
}();

var ensureIndex = function () {
    var _ref42 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee33(repository, collectionName, _ref43) {
        var fieldName = _ref43.fieldName,
            fieldType = _ref43.fieldType,
            _ref43$order = _ref43.order,
            order = _ref43$order === undefined ? 1 : _ref43$order;
        var metaCollection;
        return regeneratorRuntime.wrap(function _callee33$(_context33) {
            while (1) {
                switch (_context33.prev = _context33.next) {
                    case 0:
                        metaCollection = repository.metaCollection;

                        if (fieldName) {
                            _context33.next = 3;
                            break;
                        }

                        throw new Error('`ensureIndex` - invalid fieldName');

                    case 3:
                        if (fieldType === 'string' || fieldType === 'number') {
                            _context33.next = 5;
                            break;
                        }

                        throw new Error('`ensureIndex` - invalid fieldType');

                    case 5:
                        if (order === 1 || order === -1) {
                            _context33.next = 7;
                            break;
                        }

                        throw new Error('`ensureIndex` - invalid order type: you can use only 1 or -1');

                    case 7:
                        _context33.next = 9;
                        return count(repository, collectionName, {});

                    case 9:
                        _context33.t0 = _context33.sent;

                        if (!(_context33.t0 > 0)) {
                            _context33.next = 12;
                            break;
                        }

                        throw new Error('Can\'t create the index by \'' + fieldName + '\' field ' + ('when collection \'' + collectionName + '\' have documents'));

                    case 12:
                        _context33.next = 14;
                        return metaCollection.ensureIndex(collectionName, { fieldName: fieldName, fieldType: fieldType, order: order });

                    case 14:
                    case 'end':
                        return _context33.stop();
                }
            }
        }, _callee33, undefined);
    }));

    return function ensureIndex(_x109, _x110, _x111) {
        return _ref42.apply(this, arguments);
    };
}();

var removeIndex = function () {
    var _ref44 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee34(_ref45, collectionName, field) {
        var metaCollection = _ref45.metaCollection;
        return regeneratorRuntime.wrap(function _callee34$(_context34) {
            while (1) {
                switch (_context34.prev = _context34.next) {
                    case 0:
                        _context34.next = 2;
                        return metaCollection.removeIndex(collectionName, field);

                    case 2:
                    case 'end':
                        return _context34.stop();
                }
            }
        }, _callee34, undefined);
    }));

    return function removeIndex(_x112, _x113, _x114) {
        return _ref44.apply(this, arguments);
    };
}();

var collection = function collection(repository, collectionName) {
    return Object.freeze({
        count: count.bind(null, repository, collectionName),
        find: wrapFind.bind(null, repository, find, collectionName),
        findOne: wrapFind.bind(null, repository, findOne, collectionName),
        insert: insert.bind(null, repository, collectionName),
        remove: remove.bind(null, repository, collectionName),
        update: update.bind(null, repository, collectionName),

        ensureIndex: ensureIndex.bind(null, repository, collectionName),
        removeIndex: removeIndex.bind(null, repository, collectionName)

        // promise
        // then: array => {},
        // catch: () => {}
    });
};

exports.default = collection;
module.exports = exports['default'];